##############################################################################
import arrp_shared
import boto3
import config_arrp_global
import datetime
import db_connection
import lib_error
import logins
import random
import socket
import string
import time





##############################################################################
class rs_cluster:

  ##############################################################################
  def __init__( self, region_code, node_type_name, node_count, use_defaults, rs_parameter_group_name, rs_subnet_name, aqua_flag, postgres_flag, test_mode ):

    # MG2 : Redshift user passwords, from the docs (had to delete te last line about valid/invalid chars, non-UTF-8 characters...!)
    #       It must be 8 to 64 characters in length.
    #       It must contain at least one uppercase letter, one lowercase letter, and one number.

    # MG2 : cluster names, from the docs
    #       Must contain from 1 to 63 alphanumeric characters or hyphens.
    #       Alphabetic characters must be lowercase.
    #       First character must be a letter.
    #       Cannot end with a hyphen or contain two consecutive hyphens.

    self.__availability_zone = region_code + ('b' if region_code == 'sa-east-1' else 'a')
    self.__rs_parameter_group_name = rs_parameter_group_name
    self.__postgres_flag     = postgres_flag
    self.__rc                = boto3.client( 'redshift', region_code )
    self.__rs_subnet_name    = rs_subnet_name
    self.__test_mode         = test_mode
    self.__use_defaults      = use_defaults

    if self.__test_mode == True:
      self.cluster_aws_id = 'fetch-%s-%s-%d' % (region_code, node_type_name.replace('.','-'), node_count)
    if self.__test_mode == False:
      self.cluster_aws_id      = random.choice(string.ascii_letters) + ''.join( random.choice(string.ascii_letters + string.digits ) for i in range(0, 62) )

    self.aqua_flag           = aqua_flag
    self.bring_up_retries    = 0
    self.database            = 'dev'
    self.invalid_node_type   = False
    self.maintenance_track   = 'current'
    # self.maintenance_track   = 'trailing'
    self.most_recent_bring_up_end_utc   = None
    self.most_recent_bring_up_start_utc = None
    self.node_count          = node_count
    self.node_type_name      = node_type_name
    self.node_by_slice       = {}
    self.number_slices       = None

    if self.__test_mode == True:
      self.password            = 'matcheslightfires1A'
    if self.__test_mode == False:
      self.password            = ''.join( random.choice(string.ascii_letters + string.digits) for i in range(0, 61) ) + random.choice(string.ascii_uppercase) + random.choice(string.ascii_lowercase) + random.choice(string.digits)

    self.cluster_postgres_id = None
    self.region_code         = region_code
    self.slices_by_node      = {}
    self.username            = 'admin'
    self.version_build       = None
    self.version_major       = None
    self.version_minor       = None

    return

  ##############################################################################
  def __cluster_exist( self, error_state ):

    if error_state.is_error():
      return None, None

    available_flag = None
    exist_flag     = None

    # MG2 : took two hours to figure out how to make the exceptions work
    #       AWS docs as ever compelete, apocalyptic failures

    try:
      cluster_info = self.__rc.describe_clusters( ClusterIdentifier=self.cluster_aws_id )
    except self.__rc.exceptions.ClusterNotFoundFault:
      exist_flag = False
    except self.__rc.exceptions.InvalidTagFault:
      error_state.submit( lib_error.ERR_HARD, 'boto3', 'describe_clusters', 'Invalid request.' )
    else:
      exist_flag = True
      if cluster_info['Clusters'][0]['ClusterStatus'] == 'available':
        available_flag = True
      else:
        available_flag = False

    error_state.submit_on_error( lib_error.ERR_HARD, 'rs_cluster', '__cluster_exist', 'Failed to check if cluster exists.' )

    return available_flag, exist_flag

  ##############################################################################
  def cluster_use_or_create( self, error_state ):

    if error_state.is_error():
      return

    pg_connection_state = None

    if self.__postgres_flag == True:
      pg_connection_state = db_connection.db_connection()
      pg_connection_state.connect( error_state, config_arrp_global.PG_HOST, config_arrp_global.PG_PORT, logins.logins['postgres']['username'], config_arrp_global.PG_DB, logins.logins['postgres']['password'] )

    available_flag, exist_flag = self.__cluster_exist( error_state )

    if error_state.is_okay:
      if exist_flag == True and available_flag == True:
        self.__cluster_use( error_state, pg_connection_state )

      if exist_flag == True and available_flag == False:
        error_state.submit( lib_error.ERR_HARD, 'rs_cluster', 'cluster_use_or_create', 'Cluster "%s" exists, but is not available.' % (self.cluster_aws_id) )

      if exist_flag == False:
        self.__cluster_create( error_state, pg_connection_state )

    if self.__postgres_flag == True:
      pg_connection_state.disconnect( error_state )

    error_state.submit_on_error( lib_error.ERR_HARD, 'rs_cluster', 'cluster_use_or_create', 'Failed to create, or use existing, cluster.' )

    return

  ##############################################################################
  def __cluster_use( self, error_state, pg_connection_state ):

    if error_state.is_error():
      return

    cluster_info = self.__rc.describe_clusters( ClusterIdentifier=self.cluster_aws_id )

    self.ip   = socket.gethostbyname( cluster_info['Clusters'][0]['Endpoint']['Address'] )
    self.port = cluster_info['Clusters'][0]['Endpoint']['Port']

    if self.__postgres_flag == True:
      self.cluster_postgres_id = arrp_shared.get_existing_cluster_postgres_id( error_state, pg_connection_state, self.cluster_aws_id, self.region_code )

    if error_state.is_okay():
      rs_connection_state = db_connection.db_connection()
      rs_connection_state.connect( error_state, self.ip, self.port, self.database, self.username, self.password )
      if error_state.is_okay():
        rows, row_count = rs_connection_state.issue_sql( error_state, sql="select split_part( version(), 'Redshift ', 2 );" )
        version_major, version_minor, version_build = rows[0][0].split( '.' )
        self.version_major  = int( version_major )
        self.version_minor  = int( version_minor )
        self.version_build  = int( version_build )
        self.number_slices = 0
        rows, row_count = rs_connection_state.issue_sql( error_state, sql="select node, slice from stv_slices order by node, slice;" )
        for row in rows:
          node_id  = row['node']
          slice_id = row['slice']
          if node_id not in self.slices_by_node:
            self.slices_by_node[node_id] = []
          self.slices_by_node[node_id].append( slice_id )
          if slice_id not in self.node_by_slice:
            self.node_by_slice[slice_id] = node_id
          self.number_slices += 1

      rs_connection_state.disconnect( error_state )

    error_state.submit_on_error( lib_error.ERR_HARD, 'rs_cluster', '__cluster_use', 'Failed to use existing cluster.' )

    return

  ##############################################################################
  def __cluster_create( self, error_state, pg_connection_state ):

    if error_state.is_error():
      return

    MAX_BRING_UP_RETRIES = 5

    success_flag = False

    while error_state.is_okay() and success_flag == False and self.bring_up_retries < MAX_BRING_UP_RETRIES:
      start = datetime.datetime.now()

      try:
        self.most_recent_bring_up_start_utc = datetime.datetime.utcnow()

        if self.__use_defaults == True:
          self.__rc.create_cluster( DBName                           = self.database,
                                    ClusterIdentifier                = self.cluster_aws_id,
                                    ClusterType                      = 'multi-node',
                                    NodeType                         = self.node_type_name,
                                    MasterUsername                   = self.username,
                                    MasterUserPassword               = self.password, 
                                    AvailabilityZone                 = self.__availability_zone,
                                    ClusterParameterGroupName        = self.__rs_parameter_group_name,
                                    NumberOfNodes                    = self.node_count,
                                    PubliclyAccessible               = True,
                                    Encrypted                        = False,
                                    MaintenanceTrackName             = self.maintenance_track,
                                    # IamRoles                         = [ iam_roles ],
                                    AquaConfigurationStatus          = 'enabled' if self.aqua_flag is True else 'disabled',
                                    ManualSnapshotRetentionPeriod    = 1,
                                    AutomatedSnapshotRetentionPeriod = 1 if self.node_type_name[0:3] == 'ra3' else 0 )

        if self.__use_defaults == False:
          self.__rc.create_cluster( DBName                           = self.database,
                                    ClusterIdentifier                = self.cluster_aws_id,
                                    ClusterSubnetGroupName           = self.__rs_subnet_name,
                                    ClusterType                      = 'multi-node',
                                    NodeType                         = self.node_type_name,
                                    MasterUsername                   = self.username,
                                    MasterUserPassword               = self.password, 
                                    AvailabilityZone                 = self.__availability_zone,
                                    ClusterParameterGroupName        = self.__rs_parameter_group_name,
                                    NumberOfNodes                    = self.node_count,
                                    PubliclyAccessible               = True,
                                    Encrypted                        = False,
                                    # VpcSecurityGroupIds              = [ global_state['sg'].id ],
                                    MaintenanceTrackName             = self.maintenance_track,
                                    # IamRoles                         = [ iam_roles ],
                                    AquaConfigurationStatus          = 'enabled' if self.aqua_flag is True else 'disabled',
                                    ManualSnapshotRetentionPeriod    = 1,
                                    AutomatedSnapshotRetentionPeriod = 1 if self.node_type_name[0:3] == 'ra3' else 0 )
      except Exception as e:
        if 'Invalid node type' in str(e) or 'You do not have access' in str(e):
          self.invalid_node_type = True
        error_state.submit( lib_error.ERR_HARD, 'boto3', 'create_cluster', 'Failed to create cluster, error "%s".' % (str(e)) )

      # MG2 : the 'available' status from boto3 is completely unreliable
      #       (even for boto3 - if you have one 'available' and delete, you normally get 'operation in progress')
      #       a cluster is only available when you can DNS lookup and connect
      #       normally a cluster which comes up will become DNS/connect, but also this not always
      #       I think the slower regions (longer bring up) probably need more time for DNS/connect, as well

      if error_state.is_okay():

        if self.__postgres_flag == True:
          self.cluster_postgres_id = arrp_shared.new_cluster( error_state, pg_connection_state, self )

        error_state.print_on_error()

        self.__cluster_wait_for_status( error_state, 'available', 5, 3 )

        self.most_recent_bring_up_end_utc = datetime.datetime.utcnow() - datetime.timedelta( seconds = 5 * 3 )

        # MG2 : now, sometimes clusters say they're available, but when you try to look up their IP, or connect, you fail
        #       we also have a retry loop here, in the hope the cluster recovers
        #       but so much, since we're now paying for the cluster

        cluster_info = self.__rc.describe_clusters( ClusterIdentifier=self.cluster_aws_id )

        try:
          self.ip = socket.gethostbyname( cluster_info['Clusters'][0]['Endpoint']['Address'] )
        except Exception as e:
          error_state.submit( lib_error.ERR_HARD, 'socket', 'gethostbyname', 'Failed to get IP number for IP name "%s", error "%s".' % (cluster_info['Clusters'][0]['Endpoint']['Address'], str(e)) )
        else:
          self.port = cluster_info['Clusters'][0]['Endpoint']['Port']

        if error_state.is_okay():
          # MG2 : now check we can actually connect
          #       sometimes the cluster is up, but we can't actually connect
          #       and then we need to zap the cluster and try again with a new cluster
          rs_connection_state = db_connection.db_connection()
          rs_connection_state.connect( error_state, self.ip, self.port, self.database, self.username, self.password )
          if error_state.is_okay():
            rows, row_count = rs_connection_state.issue_sql( error_state, sql="select split_part( version(), 'Redshift ', 2 );" )
            version_major, version_minor, version_build = rows[0][0].split( '.' )
            self.version_major = int( version_major )
            self.version_minor = int( version_minor )
            self.version_build = int( version_build )
            self.number_slices = 0
            rows, row_count = rs_connection_state.issue_sql( error_state, sql="select node, slice from stv_slices order by node, slice;" )
            for row in rows:
              node_id  = row['node']
              slice_id = row['slice']
              if node_id not in self.slices_by_node:
                self.slices_by_node[node_id] = []
              self.slices_by_node[node_id].append( slice_id )
              if slice_id not in self.node_by_slice:
                self.node_by_slice[slice_id] = node_id
              self.number_slices += 1
            success_flag = True
          rs_connection_state.disconnect( error_state )

        if error_state.is_error():
          error_state.clear()
          self.cluster_issue_delete_and_wait( error_state )
          self.bring_up_retries += 1

        if self.__postgres_flag == True:
          arrp_shared.new_cluster_bring_up( error_state, pg_connection_state, self, success_flag )

    if self.bring_up_retries == MAX_BRING_UP_RETRIES:
      error_state.submit( lib_error.ERR_HARD, 'rs_cluster', '__cluster_create', 'Failed to bring up cluster after %d retries.' % (self.bring_up_retries) )

    error_state.submit_on_error( lib_error.ERR_HARD, 'rs_cluster', '__cluster_create', 'Unable to create and connect to cluster %s, bailing out.' % (self.cluster_aws_id) )

    return

  ##############################################################################
  def cluster_issue_delete_and_wait( self, error_state ):

    if error_state.is_error():
      return

    try:
      result = self.__rc.delete_cluster( ClusterIdentifier        = self.cluster_aws_id,
                                         SkipFinalClusterSnapshot = True )
    except Exception as e:
      error_state.submit( lib_error.ERR_HARD, 'boto3', 'cluster_issue_delete_and_wait', 'Failed to delete cluster.' )
    else:
      self.__cluster_wait_for_status( error_state, 'deleted', 5, 3 )

    error_state.submit_on_error( lib_error.ERR_HARD, 'rs_cluster', 'cluster_issue_delete_and_wait', 'Failed to delete cluster.' )

    return

  ##############################################################################
  def cluster_issue_delete( self, error_state ):

    if error_state.is_error():
      return

    try:
      result = self.__rc.delete_cluster( ClusterIdentifier        = self.cluster_aws_id,
                                         SkipFinalClusterSnapshot = True )
    except Exception as e:
      error_state.submit( lib_error.ERR_HARD, 'boto3', 'cluster_issue_delete', 'Failed to issue delete.' )
    else:
      self.__cluster_wait_for_status( error_state, 'deleting', 5, 1 )

    error_state.submit_on_error( lib_error.ERR_HARD, 'rs_cluster', 'cluster_issue_delete', 'Failed to issue delete to cluster.' )

    return

  ##############################################################################
  def __cluster_wait_for_status( self, error_state, status, wait_duration_in_seconds, steady_count ):

    if error_state.is_error():
      return

    print( 'cluster_wait_for_status( %s, %s )' % (self.cluster_aws_id, status) )

    status_seen_count = 0

    while status_seen_count < steady_count:
      if status != 'deleted':
        cluster_info = self.__rc.describe_clusters( ClusterIdentifier = self.cluster_aws_id )
        if 'Clusters' in cluster_info:
          print( 'cluster %s status = %s' % (self.cluster_aws_id, cluster_info['Clusters'][0]['ClusterStatus']) )
          if cluster_info['Clusters'][0]['ClusterStatus'] == status:
            status_seen_count += 1
          else:
            status_seen_count = 0
      if status == 'deleted':
        try:
          cluster_info = self.__rc.describe_clusters( ClusterIdentifier=self.cluster_aws_id )
        except Exception as e:
          status_seen_count += 1
          print( 'cluster %s status = deleted' % (self.cluster_aws_id) )
        else:
          status_seen_count = 0
          print( 'cluster %s status = %s' % (self.cluster_aws_id, cluster_info['Clusters'][0]['ClusterStatus']) )
      time.sleep( wait_duration_in_seconds )

    error_state.submit_on_error( lib_error.ERR_HARD, 'rs_cluster', '__cluster_wait_for_status', 'Failed to wait for status.' )

    return

