##############################################################################
import datetime
import db_connection
import lib_error
import rs_cluster
import rs_config
import rs_vpc
import threading
import time





##############################################################################
class rs_clusters:

  ##############################################################################
  def __init__( self, region_infos, node_type_names=['dc2.xlarge'], node_count=2, postgres_flag=True, test_mode=False ):

    self.__region_infos    = region_infos
    self.__node_type_names = node_type_names
    self.__node_count      = node_count
    self.__postgres_flag   = postgres_flag
    self.__test_mode       = test_mode

    # MG2 : region_infos = \
    #       {
    #         'region' : \
    #         {
    #           'create_config'  : True/False,
    #           'use_defaults'   : True/False,
    #           'rs_subnet_name' : String/None,
    #           'aqua_flag'      : True/False
    #         },
    #       }

    return

  ##############################################################################
  def __thread_per_region_per_node_type( self, region_code, use_defaults, rs_parameter_group_name, rs_subnet_name, aqua_flag, node_type_name, node_count, user_function, user_function_args, postgres_flag, test_mode, results ):

    error_state = lib_error.error()

    results[region_code][node_type_name] = {}

    rs_cluster_state = rs_cluster.rs_cluster( region_code, node_type_name, node_count, use_defaults, rs_parameter_group_name, rs_subnet_name, aqua_flag, postgres_flag, test_mode )

    rs_cluster_state.cluster_use_or_create( error_state )

    if rs_cluster_state.invalid_node_type == True:
      return

    rs_connection_state = db_connection.db_connection()

    if error_state.is_okay():
      rs_connection_state.connect( error_state, rs_cluster_state.ip, rs_cluster_state.port, 'dev', rs_cluster_state.username, rs_cluster_state.password )

    user_function( error_state, rs_cluster_state, rs_connection_state, results, region_code, node_type_name, node_count, user_function_args )

    error_state.submit_on_error( lib_error.ERR_HARD, 'rs_clusters', '__thread_per_region_per_node_type', 'Failed to run Redshift cluster region "%s" node type "%s".' % (region_code, node_type_name) )

    error_state.print_on_error()

    # MG2 : hack to work around lack of resource handling with errors

    error_state.clear()

    rs_connection_state.disconnect( error_state )

    if self.__test_mode == True:
      print( '*** CLUSTER DELETE IS DISABLED' )

    if self.__test_mode == False:
      error_state.clear()
      # rs_cluster_state.cluster_issue_delete( error_state )
      rs_cluster_state.cluster_issue_delete_and_wait( error_state )

    error_state.submit_on_error( lib_error.ERR_HARD, 'rs_clusters', '__thread_per_region_per_node_type', 'Failed to initialize Redshift region "%s", node type "%s".' % (region_code, node_type_name) )

    error_state.print_on_error()

    return

  ##############################################################################
  def __thread_per_region( self, region_code, region_info, node_type_names, node_count, user_function, user_function_args, postgres_flag, test_mode, results ):

    error_state = lib_error.error()

    results[region_code] = {}

    if 'create_config' in region_info and region_info['create_config'] == True:
      region_info['use_defaults']            = False
      region_info['rs_subnet_name']          = 'arrp'
      region_info['rs_parameter_group_name'] = 'arrp'
      rs_vpc_state = rs_vpc.rs_vpc( region_code, 'arrp' )
      rs_vpc_state.create( error_state )
      rs_config_state = rs_config.rs_config( region_code, rs_vpc_state, 'arrp', region_info['rs_parameter_group_name'] )
      rs_config_state.create( error_state, [5], [100] )

    # MG2 : now start one thread per node type

    if error_state.is_okay():
      thread_list = []

      for node_type_name in node_type_names:
        thread = threading.Thread( target=self.__thread_per_region_per_node_type, args=(region_code, region_info['use_defaults'], region_info['rs_parameter_group_name'], region_info['rs_subnet_name'], region_info['aqua_flag'] if 'aqua_flag' in region_info else False, node_type_name, node_count, user_function, user_function_args, postgres_flag, test_mode, results) )
        thread_list.append( thread )

      for thread in thread_list:
        thread.start()
        time.sleep( 1 )

      for thread in thread_list:
        thread.join()

    if 'create_config' in region_info and region_info['create_config'] == True:
      rs_config_state.delete( error_state )
      rs_vpc_state.delete( error_state )

    error_state.submit_on_error( lib_error.ERR_HARD, 'rs_clusters', '__thread_per_region', 'Failed to issue on Redshift region "%s".' % (region_code) )

    error_state.print_on_error()

    return

  ##############################################################################
  def run_clusters( self, error_state, user_function, user_function_args ):

    if error_state.is_error():
      return None

    results = {}
    thread_list = []

    for region_code in self.__region_infos:
      region_info = self.__region_infos[ region_code ]
      thread = threading.Thread( target=self.__thread_per_region, args=(region_code, region_info, self.__node_type_names, self.__node_count, user_function, user_function_args, self.__postgres_flag, self.__test_mode, results) )
      thread_list.append( thread )

    for thread in thread_list:
      thread.start()
      time.sleep( 1 )

    for thread in thread_list:
      thread.join()

    error_state.submit_on_error( lib_error.ERR_HARD, 'rs_clusters', 'run_clusters', 'Failed to run Redshift clusters.' )

    return results

